/**** Yet Another Alpha Processor Simulator in C++ by Kenji KISE ****/
/********************************************************************/
#include "define.h"

/*
Initialize the virtual memory space, data file looks below.
Line must begin by '@' in order to indicate the valid data.
@11ff97000 00000003
@11ff97008 1ff97138
*/

/**********************************************************************/
inline uint32_t MM_TABLE_INDEX(uint32_t addr){
  return (addr >> BLOCK_MASK_BIT);
}

inline uint32_t gen_tag(uint32_t adr){
  return (adr  & 0xffffffc0);
}

/**********************************************************************/
void memory_system::ld_inst(data_t *a, INST_TYPE *ir){
  data_t d;
  mm->ld_8byte(a, &d);
  if(a->ld()%8==0) *ir =  d.ld()      & 0xffffffff;
  if(a->ld()%8==4) *ir = (d.ld()>>32) & 0xffffffff;
}


void memory_system::ld_8byte(data_t *a, data_t *d){
  mm->ld_8byte(a, d);
}


void memory_system::st_8byte(data_t *a, data_t *d, DATA_TYPE msk){
  mm->st_8byte(a, d, msk);
}


memory_system::~memory_system(){
}


memory_system::memory_system(system_config *sc,
			     evaluation_result *ev){
  e = ev;
  mm = new main_memory(ev);

  char *prog_name = sc->program_name;

  /*** Read memory image from the file ***/
  FILE *fp;
  if((fp = fopen(prog_name, "r")) == NULL) {
    fprintf(stderr, "Bad file name: %s\n", prog_name);
    exit(1);
  }
  char buf[4096];
  fgets(buf, 4096, fp); /*** File Format Check ***/
  if(strncmp(buf+3, "SimAlpha", 8)){
    fprintf(stderr, "%s: Not a SimAlpha data file.\n", prog_name);
    exit(1);
  }
  while(!feof(fp)){
    fgets(buf, 4096, fp);
    if(*buf=='@'){
      DATA_TYPE dat;
      ADDR_TYPE adr;
      sscanf(buf+1, "%llx %llx\n", &adr, &dat);
      data_t a,d; 
      a.init(adr);
      d.init(dat);
      mm->st_8byte(&a, &d, 0);
    }
  }
  fclose(fp);
}


void memory_system::ld_nbyte(int n, data_t *a, data_t *d){
  if(a->ld()%n!=0) printf("*** ld_nbyte n=%d miss-alignment.\n", n);

  ld_8byte(a, d); /** Type Conversion **/

  int offset = a->ld() & 7;
  switch(n){
  case 1: {
    DATA_TYPE data= (d->ld() >> (offset * 8)) & 0xffllu;
    d->st(data);
    break;
  }
  case 2: {
    DATA_TYPE data= (d->ld() >> (offset * 8)) & 0xffffllu;
    d->st(data);
    break;
  }
  case 4: {
    DATA_TYPE data= (d->ld() >> (offset * 8)) & 0xffffffffllu;
    d->st(data);
    break;
  }
  case 8: {
    break;
  }
  default: printf("Case %d, Error in load_nbyte\n", n); 
    exit(1);
  }
}


void memory_system::st_nbyte(int n, data_t *a, data_t *d){
  if(a->ld()%n!=0) printf("*** st_nbyte n=%d miss-alignment.\n", n);

  int offset = a->ld() & 7;
  DATA_TYPE mask = 0;

  switch(n){ 
  case 1: {
    mask = ~(0xffllu << offset*8);
    DATA_TYPE data = (d->ld() & 0xffllu) << offset*8;
    d->st(data);
    break;
  }
  case 2: {
    mask = ~(0xffffllu << offset*8);
    DATA_TYPE data = (d->ld() & 0xffffllu) << offset*8;
    d->st(data);
    break;
  }
  case 4: {
    mask = ~(0xffffffffllu << offset*8);
    DATA_TYPE data = (d->ld() & 0xffffffffllu) << offset*8;
    d->st(data);
    break;
  }
  case 8: {
    mask = 0;
    break;
  }
  default: printf("Case %d, Error in store_nbyte\n", n); 
    exit(1);
  }

  st_8byte(a, d, mask); /** Type Conversion **/
}

/**********************************************************************/
main_memory::main_memory(evaluation_result *ev){
  e = ev;
  for(int i=0; i<BLOCK_TABLE_SIZE; i++) block_table[i]=NULL;
}

data_t *main_memory::allocblock(data_t *a){
  data_t *ret=(data_t *)calloc(BLOCK_SIZE/DATA_T_SIZE,
                               sizeof(class data_t));
  block_table[MM_TABLE_INDEX(a->ld())]=ret;
  if(ret==NULL){
    printf("** Error: memory allocation in allocblock.\n");
    exit(0);
  }
  e->used_memory_block++;
  return ret;
}

void main_memory::ld_8byte(data_t *a, data_t *d){
  ADDR_TYPE adr = a->ld() & ~7;
  data_t *ptr = block_table[MM_TABLE_INDEX(adr)];
  unsigned int offset = (adr & BLOCK_MASK)/DATA_T_SIZE;
  if(ptr==NULL) ptr=allocblock(a);
  *d = *(ptr + offset); /** COPY **/
}

void main_memory::st_8byte(data_t *a, data_t *d, DATA_TYPE msk){
  ADDR_TYPE adr = a->ld() & ~7;
  data_t *ptr = block_table[MM_TABLE_INDEX(adr)];
  unsigned int offset = (adr & BLOCK_MASK)/DATA_T_SIZE;
  if(ptr==NULL) ptr=allocblock(a);
  (ptr + offset)->st( ((ptr + offset)->ld() & msk) | d->ld() );
}
/*********************************************************************/
